/* UpTools v8.6
 *
 * Copyright (c) 2005-2020 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <UpTools/UpRoute2.h>
#include <UpTools/UpLog.h>
#include <cassert>


UpRoute2::UpRoute2()
: metric(0),parent(0),order(0),maskLen(0),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const UpRoute2& route)
: UpNet(route.ip,route.m),metric(route.metric),parent(0),order(0),maskLen(route.maskLen),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const UpAddress& baseAddr,const UpAddress& mask,unsigned int _metric)
: UpNet(baseAddr,mask),metric(_metric),parent(0),order(0),maskLen(m.maskBitsLen()),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const UpAddress& baseAddr,unsigned int _maskLen,unsigned int _metric)
: UpNet(baseAddr,_maskLen),metric(_metric),parent(0),order(0),maskLen(_maskLen),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::UpRoute2(in_addr baseAddr,in_addr mask,unsigned int _metric)
: UpNet(baseAddr,mask),metric(_metric),parent(0),order(0),maskLen(m.maskBitsLen()),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const in6_addr& baseAddr,const in6_addr& mask,unsigned int _metric)
: UpNet(baseAddr,mask),metric(_metric),parent(0),order(0),maskLen(m.maskBitsLen()),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const std::string& route)
: metric(0),parent(0),order(0),auxNet(false)
{ set(route); links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const std::string& route,unsigned int _metric)
: metric(_metric),parent(0),order(0),auxNet(false)
{ set(route,_metric); links[0] = links[1] = 0; }

UpRoute2::UpRoute2(const UpNet& n,unsigned int _metric)
: UpNet(n), metric(_metric),parent(0),order(0), maskLen(m.maskBitsLen()),auxNet(false)
{ links[0] = links[1] = 0; }

UpRoute2::~UpRoute2() {
	delete links[0];
	delete links[1];
}

bool UpRoute2::set(const std::string& net,unsigned int _metric) {
	size_t pos=net.find('/');
	if( pos!=std::string::npos ) { // slash found
		ip=net.substr(0,pos++);
		if( net.find_first_of(".:",pos)!=std::string::npos ) {
			m = net.c_str()+pos; maskLen = m.maskBitsLen();
		} else {
			int len = std::atoi(net.c_str()+pos);
			if( len>=0 ) { m.setMask((unsigned int)len,ip.addrType); maskLen = len; }
			else { m.clear(); maskLen = 0; }
		}
	} else { // slash not found: one-host net
		ip=net;
		m.setMask(128,ip.addrType); maskLen = 128;
	}
	metric = _metric;
	if( ip.family()!=m.family() or ! (m.family()==AF_INET or m.family()==AF_INET6) )
	{ ip.clear(); m.clear(); maskLen = 0; return false; }
	
	return true;
}

bool UpRoute2::set(const std::string& net) {
	size_t pos = net.find('/'), spc;
	if( pos!=std::string::npos ) { // slash found
		ip = net.substr(0,pos++);
		spc = net.find(' ',pos);
		if( net.find_first_of(".:",pos)!=std::string::npos ) { // mask as addr
			if( spc!=std::string::npos ) { // metric found?
				m = net.substr(pos,spc-pos);
				metric = (spc<net.size()) ? std::atoi(net.c_str()+spc+1) : 0;
			} else { // no metric found
				m = net.c_str()+pos;
				metric = 0;
			}
			maskLen = m.maskBitsLen();
		} else { // mask as number
			int len = std::atoi(net.c_str()+pos);
			if( len>=0 ) { m.setMask((unsigned int)len,ip.addrType); maskLen = len; }
			else { m.clear(); maskLen = 0; }
			metric = (spc!=std::string::npos) ? std::atoi(net.c_str()+spc+1) : 0;
		}
	} else { // slash not found: one-host net
		spc =  net.find(' ');
		if( spc!=std::string::npos ) { // metric found?
			ip = net.substr(0,spc++);
			metric = std::atoi(net.c_str()+spc);
		} else {
			ip = net;
			metric = 0;
		}
		m.setMask(128,ip.addrType); maskLen = 128;
	}
	if( ip.family()!=m.family() or ! (m.family()==AF_INET or m.family()==AF_INET6) )
	{ ip.clear(); m.clear(); maskLen = 0; return false; }
	return true;
}

//////////////////////////////////////////////////////////////////////////////
	

void UpRoute2::becomeAncestor(UpRoute2* r1,UpRoute2* r2) {
	if( r1->maskLen < r2->maskLen)	{ m = r1->m; maskLen = r1->maskLen; }
	else 										{ m = r2->m; maskLen = r2->maskLen; }
	while( (r1->ip & m) != (r2->ip & m) ) { m += m; --maskLen; }
	// it found a common parent
	ip = (r1->ip & m);
}

UpNetRelation UpRoute2::addCheck(UpRoute2* r) {
	UpNetRelation incl = check(*r);
	if( incl!=UPNR_SUBNET ) return incl;
	bool neworder = r->ip.getBit(((family()==AF_INET)?31:127)-maskLen);
	UpRoute2*& link = links[neworder];
	if( !link ) {
		r->parent = this;
		r->order = neworder;
		link = r;
		return UPNR_SUBNET;
	}
	switch( incl=link->addCheck(r) ) {
		case UPNR_SUPERNET: {
			UpRoute2* q = link;
			r->parent = this;
			r->order = neworder;
			r->addCheck(q);
			link = r;
			return UPNR_SUBNET;
		}
		case UPNR_BEFORE:
		case UPNR_AFTER: {
			UpRoute2* q = new UpRoute2;
			q->becomeAncestor(link,r);
			q->auxNet = true;
			if( q->maskLen==maskLen ) {
				UPLOG(LOG_ERR,"Error in UpRoute2::addCheck(r)");
				delete q;
				return UPNR_SUBNET_COLLIDES;
			}
			q->parent = this;
			q->order = neworder;
			q->addCheck(r);
			q->addCheck(link);
			link = q;
			return UPNR_SUBNET;
		}
		case UPNR_EQUAL:
			if( link->auxNet ) { // we can replace it!!
				r->parent = this;
				r->order = neworder;
				if( link->links[0] ) {
					r->links[0] = link->links[0];
					r->links[0]->parent = r; // and it keeps its order
					link->links[0] = 0;
				}
				if( link->links[1] ) {
					r->links[1] = link->links[1];
					r->links[1]->parent = r; // and it keeps its order
					link->links[1] = 0;
				}
				delete link;
				link = r;
				return UPNR_SUBNET;
			}
			return UPNR_SUBNET_COLLIDES;

		case UPNR_SUBNET_COLLIDES:
		case UPNR_INVALID_RELATION:
		case UPNR_SUBNET:
		default: // this default here is just to turn off compile warnings
			return incl;
	}
	return UPNR_INVALID_RELATION;
}


// removes a net and possibly its subnets
// returns true if it removes something
bool UpRoute2::removeTest(const UpAddress& _ip,unsigned char _maskLen,bool includeSubnets) {
	if( check(_ip)==UPNR_SUBNET ) { // this route contains _ip
		if( maskLen==_maskLen ) {
			// this route is equal to requested route, and should be deleted
			// the parent of this route will check if there is need to save subnets
			return true;
		}
		if( maskLen < _maskLen ) { // this route is larger than requested route, and contains it
			for( unsigned char ord=0; ord<2; ++ord) {
				UpRoute2*& link(links[ord]);
				if( link and link->removeTest(_ip,_maskLen,includeSubnets) ) {
					UpRoute2* toDel = link; // the route pointed by link should be deleted
					if( !includeSubnets ) {
						if( toDel->links[0] ) {
							link = toDel->links[0];
							toDel->links[0] = 0;
							link->parent = this;
							link->order = ord;
							if( toDel->links[1] ) {
								addCheck(toDel->links[1]);
								toDel->links[1] = 0;
							}
						} else if( toDel->links[1] ) {
							link = toDel->links[1];
							toDel->links[1] = 0;
							link->parent = this;
							link->order = ord;
						} else {
							link = 0; // prevent this->links[ord] to continue pointing to a deleted UpRoute2
						}
					} else {
						link = 0; // prevent this->links[ord] to continue pointing to a deleted UpRoute2
					}
					delete toDel;
				}
			}
		} else { // this route is smaller than requested route, and is contained by it
			if( includeSubnets ) return true;
		}
	}
	if( auxNet and links[0]==0 and links[1]==0 ) return true; // this route should be deleted
	return false; // this route should not be deleted
}


void UpRoute2::clear() {
	ip.clear(); m.clear(); auxNet = false; metric = 0; maskLen = 0;
	if( links[0] ) { delete links[0]; links[0] = 0; }
	if( links[1] ) { delete links[1]; links[1] = 0; }
	parent = 0;
}

//////////////////////////////////////////////////////////////////////////////

void UpRoute2::tree(std::ostream& o, int indent) const
{
	if( !auxNet )  {
		for(int i=0;i<indent;++i) o.write("  ",2);
		append(o);
		if( auxNet ) o << "(aux)\n"; else o << '\n';
  		++indent;
	}
	if(links[0]) links[0]->tree(o,indent);
	if(links[1]) links[1]->tree(o,indent);
}

void UpRoute2::trie(std::ostream& o, int indent) const
{
	for(int i=0;i<indent;++i) o.write("  ",2);
	append(o);
	if( auxNet ) o << "(aux)\n"; else o << '\n';
  	++indent;
	if(links[0]) links[0]->trie(o,indent);
	if(links[1]) links[1]->trie(o,indent);
}


//////////////////////////////////////////////////////////////////////////////

const UpRoute2* UpRoute2::lookup(in_addr ipv4,bool useMetric) const {
	UpNetRelation incl;
	const UpRoute2 *p=this,*ret=0;
	do {
		incl = p->check(ipv4);
		if( incl==UPNR_SUBNET ) {
			if( !p->auxNet and (!useMetric or !ret or p->metric<=ret->metric) )
				ret = p;
			p = p->links[0] ? p->links[0] : p->links[1];
		} else if( incl==UPNR_AFTER ) {
			if( p->order==0 and p->parent ) p = p->parent->links[1]; else break;
		} else break;
	} while(p);
	return ret;
}


const UpRoute2* UpRoute2::lookup(const in6_addr& ipv6,bool useMetric) const {
	UpNetRelation incl;
	const UpRoute2 *p=this,*ret=0;
	do {
		incl = p->check(ipv6);
		if( incl==UPNR_SUBNET ) {
			if( !p->auxNet and (!useMetric or !ret or p->metric<=ret->metric) )
				ret = p;
			p = p->links[0] ? p->links[0] : p->links[1];
		} else if( incl==UPNR_AFTER ) {
			if( p->order==0 and p->parent ) p = p->parent->links[1]; else break;
		} else break;
	} while(p);
	return ret;
}

const UpRoute2* UpRoute2::lookup(const UpAddress& addr,unsigned int _maskLen,bool useMetric) const {
	UpNetRelation incl;
	const UpRoute2 *p=this,*ret=0;
	do {
		incl = p->check(addr);
		if( incl==UPNR_SUBNET ) {
			if( p->maskLen > _maskLen ) break; // n is a bigger
			if( !p->auxNet and (!useMetric or !ret or p->metric<=ret->metric) )
				ret = p;
			p = p->links[0] ? p->links[0] : p->links[1];
		} else if( incl==UPNR_AFTER ) {
			if( p->order==0 and p->parent ) p = p->parent->links[1]; else break;
		} else break;
	} while(p);
	return ret;
}


std::list<const UpRoute2*> UpRoute2::lookupAll(const UpAddress& ipAddr) const {
	UpNetRelation incl;
	const UpRoute2 *p=this;
	std::list<const UpRoute2*> ret;
	do {
		incl = p->check(ipAddr);
		if( incl==UPNR_SUBNET ) {
			if( ! p->auxNet ) ret.push_back(p);
			p = p->links[0] ? p->links[0] : p->links[1];
		} else if( incl==UPNR_AFTER ) {
			if( p->order==0 and p->parent ) p = p->parent->links[1]; else break;
		} else break;
	} while(p);
	return ret;
}
