/* UpTools v8.6
 *
 * Copyright (c) 2005-2020 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpSqlMysql.h>
#include <UpTools/UpLog.h>
#include <UpTools/UpArch.h>
#include <mysql/errmsg.h>

void UpSqlConnMysql::disconnect() {
	if(need_close) {
		mysql_close(&conn);
		need_close = false;
		statusOk = false;
		statusPhrase = NOT_CONNECTED;
	}
}


bool UpSqlConnMysql::privateConnect(const std::string& host)
{
	if( 0 == mysql_init(&conn) ) {
		statusOk = false;
		statusPhrase = mysql_error(&conn);
		UPLOG(LOG_ERR,"UpSqlConnMysql::connect() init error: %s",statusPhrase);
		return false;
	}
	need_close = true;
	unsigned int conn_timeout=UPMYSQL_CONNECT_TIMEOUT_;
	mysql_options(&conn,MYSQL_OPT_CONNECT_TIMEOUT,(const char*)&conn_timeout);
	#ifdef CLIENT_MULTI_STATEMENTS
		#define MYCLIENT_FLAGS CLIENT_MULTI_STATEMENTS
	#else
		#define MYCLIENT_FLAGS 0
	#endif
	scrPass();
	mysql_real_connect( &conn, host.c_str(), user.c_str(), pass.c_str(),
							  dbname.c_str(), port, 0, MYCLIENT_FLAGS  );
	scrPass();
	statusOk = (mysql_errno(&conn)==0);
	if( statusOk ) {
		statusPhrase = "ok";
		// UPLOG(LOG_INFO,"UpSqlConnMysql::connect() connected to %s",host.c_str());
	} else {
		statusPhrase = mysql_error(&conn);
		UPLOG(LOG_ERR,"UpSqlConnMysql::connect(host) failed connection to %s: %s",
				host.c_str(), statusPhrase );
	}
	return statusOk;
}



bool UpSqlConnMysql::exec(UpSqlQuery& q,const char* sqlText,size_t sqlTextLen) {
	// ojo que connect() no borre este DbElem
	if( !statusOk and !connect() )
		{ q.setError(CR_SERVER_LOST,statusPhrase); return false; }

	if( mysql_real_query(&conn, sqlText, sqlTextLen ) != 0 ) {
		// query fails for the first time (not necessarily a connection error)
		statusPhrase = mysql_error(&conn); 
		unsigned int errNum = mysql_errno(&conn);
		if( !errorIsConnectionError(errNum) ) // query error
			{ q.setError(errNum,statusPhrase); return false; } // but conn still ok
		// connection error, try to reconnect
		UPLOG(LOG_ERR,"UpSqlConnMysql::exec() %s, trying to reconnect",statusPhrase);
		if( !connect() ) // reconnect failed, statusOk has been set to 0 in connect()
			{ q.setError(errNum,statusPhrase); return false; }
		// try to query again after reconnection
		if( mysql_real_query(&conn, sqlText, sqlTextLen ) != 0 ) {
			// query fails for the second time, no reconnection this time
			statusPhrase = mysql_error(&conn); 
			errNum = mysql_errno(&conn);
			q.setError(errNum,statusPhrase);
			// we only update connection status if it is a connection error
			if( errorIsConnectionError(errNum) ) statusOk = false;
			return false;
		}
		// go on and process query results
	}
	// query succeded
	statusPhrase = "ok";
	statusOk = true;
	q.clearResultList();
	do {
		MYSQL_RES* res;
		unsigned int fieldCount=mysql_field_count(&conn);
		if( fieldCount!=0 and (res=mysql_store_result(&conn))!=0 ) {
			q.addQueryElem(
				new QueryElemMysql( res , fieldCount , mysql_insert_id(&conn), &conn )
			);
		} else {
			// (fieldCount==0) or query error => query should not
			// return data (UPDATE, INSERT or DELETE)
			q.addQueryElem(
				new QueryElemMysql( 0 , 0 , mysql_insert_id(&conn) , &conn )
			);
		}
	} while( mysql_next_result(&conn)==0 or mysql_errno(&conn) );
	return true;
}


bool UpSqlConnMysql::errorIsConnectionError(unsigned int err) {
	return 
		err==CR_UNKNOWN_ERROR or
		err==CR_SOCKET_CREATE_ERROR or
		err==CR_CONNECTION_ERROR or
		err==CR_CONN_HOST_ERROR or
		err==CR_IPSOCK_ERROR or
		err==CR_UNKNOWN_HOST or
		err==CR_SERVER_GONE_ERROR or
		err==CR_VERSION_ERROR or
		err==CR_OUT_OF_MEMORY or
		err==CR_WRONG_HOST_INFO or
		err==CR_SERVER_HANDSHAKE_ERR or
		err==CR_SERVER_LOST or
		err==CR_NAMEDPIPEWAIT_ERROR or
		err==CR_NAMEDPIPEOPEN_ERROR or
		err==CR_NAMEDPIPESETSTATE_ERROR or
		err==CR_SSL_CONNECTION_ERROR;
}
