#include <UpTools/UpTimeVal.h>
#include <UpTools/UpInet.h>
#include <UpTools/UpRoute2.h>
#include <UpTools/UpRoute.h>
#include <UpTools/UpConf.h>

#include <iostream>
#include <fstream>

using namespace std;

// UpRoute2 is a class for very big ip routing tables.
// It has similar lookup times as UpRoute, but it has several orders of magnitude
// faster insertion in big tables!!! and needs no optimization() step!!!.
// Internally it is implemented as a PATRICIA trie variant.

///////////////////////////////////////////////////////////////////////////////////////
class Route2 : public UpRoute2 {
	private:
		// UpAddresses 'ip' and 'm', and unsigned int 'metric' are
		// inherited from UpNet
		UpAddress gateway;
	public:
		Route2(unsigned int family)
		{
			if(      family==AF_INET  ) { m = ip = gateway = "0.0.0.0"; }
			else if( family==AF_INET6 ) { m = ip = gateway = "::0"; }
		}
		virtual	~Route2()
		{}

		Route2(const string& s) // format: "ipAddr/maskLen cost gatewayAddr"
		{ setData(s); }

		bool setData(const string& s)
		{
			size_t pos( s.rfind(' ') ); if( pos==string::npos ) { cleanData(); return false; }
			set(s.substr(0,pos++)); // UpRoute2::set(const std::string&)
									// this sets ip/mask and metric
			if( pos>=s.size() ) { cleanData(); return false; }
			gateway.setf(s.c_str()+pos,ip.family());
			if( gateway.family()!=ip.family() ) { cleanData(); return false; }
			return true;
		}

		virtual void cleanData()
		{ clear(); gateway.clear(); }

		virtual void append(ostream& o) const
		{ o<<ip<<'/'<<getMaskLen()<<' '<<metric<<' '<<gateway; }

};


class Route : public UpRoute {
	private:
		// UpAddresses 'ip' and 'm', and unsigned int 'metric' are
		// inherited from UpNet
		UpAddress gateway;
	public:
		Route(unsigned int family)
		{
			if(      family==AF_INET  ) { m = ip = gateway = "0.0.0.0"; }
			else if( family==AF_INET6 ) { m = ip = gateway = "::0"; }
		}
		virtual	~Route()
		{}

		Route(const string& s) // format: "ipAddr/maskLen cost gatewayAddr"
		{ setData(s); }

		bool setData(const string& s)
		{
			size_t pos( s.rfind(' ') ); if( pos==string::npos ) { cleanData(); return false; }
			set(s.substr(0,pos++)); // UpRoute2::set(const std::string&)
									// this sets ip/mask and metric
			if( pos>=s.size() ) { cleanData(); return false; }
			gateway.setf(s.c_str()+pos,ip.family());
			if( gateway.family()!=ip.family() ) { cleanData(); return false; }
			return true;
		}

		virtual void cleanData()
		{ clear(); gateway.clear(); }

		virtual void append(ostream& o) const
		{ o<<ip<<'/'<<getMaskLen()<<' '<<metric<<' '<<gateway; }

};


///////////////////////////////////////////////////////////////////////////////////////

// forward declarations
void randomizeInNet(UpAddress& upa,UpNet& r);
int signedRand();


///////////////////////////////////////////////////////////////////////////////////////


int main(int argc, char* argv[]) {

	if(argc!=4) {
		cerr << "Group lookup time test\n"
		<< "Usage: " << argv[0] << " routes-file test-net sample-count\n"
		<< "routes are parsed from the route-file\n"
		<< "sample-count random ip's will be generated in the test-net\n"
		<< "test-net format is ip/mask\n"
		<< "test-net and routes-file must be have the same ip version (ipv4 or ipv6)"
		<< endl;
		return 1;
	}

	unsigned int sampleCount = atoi( argv[3] );

	/////////////////////////////////////////////////////////////////////
	UpNet testNet(argv[2]); // samples will be generated in this net
	if( testNet.family()==AF_UNSPEC ) {
		cerr << "Bad test-net: " << argv[2] <<endl;
		return 2;
	}
	/////////////////////////////////////////////////////////////////////
	// route table
	Route rootRoute( testNet.family() ); // 0.0.0.0/0 or ::0/0
	Route2 rootRoute2( testNet.family() ); // 0.0.0.0/0 or ::0/0
	// most expensive metric for default route
	rootRoute.setMetric((unsigned int)(-1));
	rootRoute2.setMetric((unsigned int)(-1));

	// true: address lookups return longest prefix network
	// false: address lookups will return minimum metric networks
	bool minMetric = true;

	/////////////////////////////////////////////////////////////////////
	// configuration object
	UpConf conf;
	if( !conf.parseFile(argv[1]) ) {
		cerr << "Bad routes-file " << argv[1]
		<< " : " << conf.getStatusPhrase() << endl;
		return 3;
	}

	/////////////////////////////////////////////////////////////////////
	// generate sample addresses inside testNet
	cout << "Generating " << sampleCount << " sample addresses inside network "
		 << testNet << " ..." << endl;
	UpAddress* testAddr = new UpAddress[sampleCount];
	UpAddress* lastAddr(testAddr+sampleCount); // pointer to end of table
	UpAddress* addrPtr;
	for(addrPtr = testAddr; addrPtr!=lastAddr; ++addrPtr) {
		// set UpAddress (*addrPtr) to ba a random address belonging to testNet
		randomizeInNet(*addrPtr,testNet);
	}

	cout << "done.\n\n" <<  endl;

	/////////////////////////////////////////////////////////////////////
	// load table in rootRoute
	cout << "Loading routes from file \"" << argv[1] << "\" ..." << endl;
	string routeLine;
	unsigned int routeCount=0;
	UpTimeVal tt; tt.tic();
	while( (routeLine=conf("routes")).size() ) {
		Route* r = new Route( routeLine ); // create a new route
		UpNetRelation i = rootRoute.addCheck( r ); // add new route to root
		// cout<<"Adding: "<<*r<<" -------------"<<UpNet::relationString(i)<<endl;
		if( i==UPNR_SUBNET ) { ++routeCount; continue; }

		if( i==UPNR_INVALID_RELATION ) {
			cerr << "Bad route in route-file: " << routeLine <<endl;
		} else if( i==UPNR_EQUAL ) {
			rootRoute.setData( routeLine ); // sets default gateway
			++routeCount;
		} else if( i==UPNR_SUBNET_COLLIDES ) {
			cout<<"repeated route: "<<routeLine<<endl;
		} else {
			cerr<<"strange route("<<UpNet::relationString(i)<<"): "<<routeLine<<endl;
		}
		delete r;
	}
	tt.toc();
	cout << "Loaded " << routeCount << " routes in "<<tt<<" secs\n\n" << endl;


	/////////////////////////////////////////////////////////////////////
	// load table in rootRoute2
	conf("resetRead");
	cout << "Loading routes2 from file \"" << argv[1] << "\" ..." << endl;
	routeCount=0;
	tt.tic();
	while( (routeLine=conf("routes")).size() ) {
		Route2* r = new Route2( routeLine ); // create a new route
		UpNetRelation i = rootRoute2.addCheck( r ); // add new route to root
		// cout<<"Adding: "<<*r<<" -------------"<<UpNet::relationString(i)<<endl;
		if( i==UPNR_SUBNET ) { ++routeCount; continue; }

		if( i==UPNR_INVALID_RELATION ) {
			cerr << "Bad route2 in route-file: " << routeLine <<endl;
		} else if( i==UPNR_EQUAL ) {
			rootRoute2.setData( routeLine ); // sets default gateway
			++routeCount;
		} else if( i==UPNR_SUBNET_COLLIDES ) {
			cout<<"repeated route2: "<<routeLine<<endl;
		} else {
			cerr<<"strange route2("<<UpNet::relationString(i)<<"): "<<routeLine<<endl;
		}
		delete r;
	}
	tt.toc();
	cout << "Loaded2 " << routeCount << " routes in "<<tt<<" secs\n\n" << endl;


	/////////////////////////////////////////////////////////////////////
	// test remove some routes
	cout<<"\nRemoving some routes\n";
	UpNet upn;
	tt.tic();
	while( (routeLine=conf("remove")).size() ) {
		std::string sn(routeLine);
		upn.set(sn.substr(0,sn.find(' ')));
		// cout<<"Removing "<<upn<<'\n';
		rootRoute.remove(upn.getBase(),upn.getMask(),false);
	}
	tt.toc();
	cout<<"Removal took "<<tt<<" secs\n";

	// print route tree
	ofstream fr1("/tmp/fr1");
	cout << "Printing route tree to /tmp/fr1\n";
	rootRoute.tree(fr1);
	fr1.close();

	/////////////////////////////////////////////////////////////////////
	// test remove some routes
	cout<<"\nRemoving some routes2\n";
	conf("resetRead");
	tt.tic();
	while( (routeLine=conf("remove")).size() ) {
		std::string sn(routeLine);
		upn.set(sn.substr(0,sn.find(' ')));
		// cout<<"Removing "<<upn<<'\n';
		rootRoute2.remove(upn.getBase(),upn.getMask(),false);
	}
	tt.toc();
	cout<<"Removal took "<<tt<<" secs\n";

	// print route tree
	ofstream fr2("/tmp/fr2");
	cout << "Printing route2 tree to /tmp/fr2\n";
	rootRoute2.tree(fr2);
	fr2.close();

	/////////////////////////////////////////////////////////////////////
	// optimize route tree and print
	cout << "\n\nOptimizing route tree..." << endl;
	tt.tic();
		rootRoute.optimize();
	tt.toc();
	cout << "Optimization time: " << tt << " secs"<<endl;
	/////////////////////////////////////////////////////////////////////
	// search for sample addresses in table
	cout << "\nLooking up " << sampleCount << " addresses in table..." << endl;
	UpTimeVal tunop;
	unsigned int errors = 0;
	tunop.tic();
	for(addrPtr = testAddr; addrPtr!=lastAddr; ++addrPtr) {
		const UpRoute* r = rootRoute.lookup(*addrPtr,minMetric);
		const UpRoute2* r2 = rootRoute2.lookup(*addrPtr,minMetric);

		if( r->getBase()!=r2->getBase() or r->getMask()!=r2->getMask() ) {
			cout<<"Lookup mismatch! addr="<<*addrPtr<<"   (r="<<r->getBase()<<'/'<<r->getMask().maskBitsLen()<<") != (r2="
					<<r2->getBase()<<'/'<<r2->getMaskLen()<<")"<<endl;
			++errors;
			// re lookup:
			r2 = rootRoute2.lookup(*addrPtr,minMetric);
			cout<<"Re Lookup "<<*addrPtr<<"  r2="<<r2->getBase()<<'/'<<r2->getMaskLen()<<endl;
		} else {
			// cout<<"Lookup ok, lookup("<<*addrPtr<<") = "<<r2->getBase()<<'/'<<r2->getMaskLen()<<endl;
		}
	}
	tunop.toc();
	cout << "Lookup of " << sampleCount <<" addresses lasted " << tunop
	<< "secs = " << (tunop/(double)sampleCount) << " secs/addr"<< endl;
	cout<<"Lookup "<<sampleCount<<" resulted in "<<errors<<" errors"<<endl;

	delete[] testAddr;

	return 0;
}





///////////////////////////////////////////////////////////////////////////////////////
// this functions generates a random addres belonging to network r
// and stores it in upa
void randomizeInNet(UpAddress& upa,UpNet& r) {
	UpAddress randAddr;
	if( r.family()==AF_INET ) {
		union {
			int i;
			in_addr a4; // in_addr defined in /usr/include/netinet/in.h
		} u4;
		u4.i = signedRand();
		randAddr = u4.a4; // UpAddress::operator=(const in6_addr&)
	} else if( r.family()==AF_INET6 ) {
		union {
			struct { int a,b,c,d; } ints;
			in6_addr a6; // in6_addr defined in /usr/include/netinet/in.h
		} u6;
		u6.ints.a = signedRand();
		u6.ints.b = signedRand();
		u6.ints.c = signedRand();
		u6.ints.d = signedRand();
		randAddr = u6.a6; // UpAddress::operator=(const in6_addr&)
	} else return;
	UpAddress m = r.getMask();
	m.negate();
	upa = r.getBase() + ( randAddr & m );
}


///////////////////////////////////////////////////////////////////////////////////////
// this functions generates a random number between -2147483647 and 2147483647
// (not too random, but it's ok)
int signedRand() {
	static int z = 1;
	z = -z;
	return rand() * z;
}

