/* UpTools v8.6
 *
 * Copyright (c) 2005-2020 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UPROUTE2_H
#define UPROUTE2_H

#include <UpTools/UpInet.h>
#include <list>

// UpRoute2 is a class for very big ip routing tables.
// It has similar lookup times as UpRoute, but it has several orders of magnitude
// faster insertion in big tables!!! and needs no optimization() step!!!.
// Internally it is implemented as a PATRICIA trie variant.

class UpRoute2 : public UpNet {
public:

	UpRoute2();
	UpRoute2(const UpRoute2&);
	UpRoute2(const UpNet&,unsigned int _metric=0);
	UpRoute2(const UpAddress& baseAddr,const UpAddress& mask,unsigned int _metric=0);
	UpRoute2(const UpAddress& baseAddr,unsigned int _maskLen,unsigned int _metric=0);
	UpRoute2(in_addr baseAddr,in_addr mask,unsigned int _metric=0);   // net order
	UpRoute2(const in6_addr& baseAddr,const in6_addr& mask,unsigned int _metric=0);   // net order
	UpRoute2(const std::string& net);
	UpRoute2(const std::string& net,unsigned int _metric);
	virtual ~UpRoute2();

	// This class inherits getBase(), getMask()
	// getLastIp(), getSize(), all check(...) and all includes(...) methods

	// clear internal data
	virtual void clear(); // clear internal data
	// set internal data
	virtual void set(const UpRoute2&);
	virtual void set(const UpNet&,unsigned int _metric=0);
	virtual void set(const UpAddress& baseAddr,const UpAddress& mask,unsigned int _metric=0);
	virtual void set(const UpAddress& baseAddr,unsigned int _maskLen,unsigned int _metric=0);
	virtual void set(in_addr baseAddr,in_addr mask,unsigned int _metric=0);   // net order
	virtual void set(const in6_addr& baseAddr,const in6_addr& mask,unsigned int _metric=0);   // net order
	virtual bool set(const std::string& net,unsigned int _metric);
	virtual bool set(const std::string& net);
	void setMetric(unsigned int _metric);

	unsigned int getMaskLen() const;

	bool operator<(const UpRoute2&);

	UpRoute2&         operator=(const UpRoute2&);
	virtual UpRoute2& operator=(const std::string&); // same as set(...)
	// examples: "200.64.0.0/10 5" or "200.64.0.0/255.192.0.0 8"
	// examples: "2001::/16 20"     or "2001::/ffff:: 0"

	// Add an UpRoute2 to internal network tree. The route being added must be a
	// subnet of (*this) in order to be accepted.
	// If it is accepted then it returns UPNR_SUBNET, and the acceptor stays in
	// charge of deleting the added UpRoute2.
	UpNetRelation addCheck(UpRoute2* r);
	// Remove an UpRoute2 from internal network tree, and optionally its subnets.
	void remove(const UpAddress& _ip,const UpAddress& _m,bool includeSubnets=false);
	void remove(const UpAddress& _ip,unsigned int _maskLen,bool includeSubnets=false);
	// Return a pointer to the most specific UpRoute2 including addr,
	// or 0 if no UpRoute2 is found that includes addr.
	// If useMetric is true, then the UpRoute2 with the smaller metric including
	// addr is returned instead.
	const UpRoute2* lookup(const UpAddress& addr,bool useMetric=false) const;
	const UpRoute2* lookup(const UpSockAddr& sa ,bool useMetric=false) const;
	const UpRoute2* lookup(in_addr addr         ,bool useMetric=false) const;
	const UpRoute2* lookup(const in6_addr& addr ,bool useMetric=false) const;

	const UpRoute2* lookup(const UpNet& net     ,bool useMetric=false) const;
	const UpRoute2* lookup(const UpAddress& net,unsigned int _maskLen,bool useMetric=false) const;
	// Return all the UpRoute2s in the tree that contain addr
	// (the user may later choose the route he likes)
	std::list<const UpRoute2*> lookupAll(const UpAddress& addr) const;
	// show internal network tree including auxiliary nodes
	void trie(std::ostream& o, int indent=0) const;
	// show internal network tree
	void tree(std::ostream& o, int indent=0) const;
	// get next node in tree
	const UpRoute2* getNext() const;
	UpRoute2* getNext();

	virtual void append(std::ostream& o) const;
	virtual void append(std::string& o) const;
	friend std::ostream& operator<<(std::ostream& o,const UpRoute2& net);
	friend std::string&  operator<<(std::string& o,const UpRoute2& net);

protected:
	//	UpAddress ip and m are inherited from UpNet
	unsigned int metric;
	UpRoute2 *links[2],*parent;
	unsigned char order;
	unsigned char maskLen;
	bool auxNet;

private:
	void becomeAncestor(UpRoute2*,UpRoute2*);
	UpRoute2* getNextPriv();
	UpRoute2* getNextPriv(unsigned char ord);
	bool removeTest(const UpAddress& _ip,unsigned char _maskLen,bool includeSubnets);

	
};



#endif

#include <UpTools/UpRoute2.inl>
